# Copyright (c) 2022 DomainTools LLC
# Copyright (c) 2015-2016, 2019, 2021 by Farsight Security, Inc.
# This work is unpublished, and unauthorized reproduction is prohibited.
# Farsight Security, Inc. owns all rights to this work and intends that
# it be held confidential, preserved as a trade secret, and protected as
# an unpublished copyrighted work.

import errno
import http.client
import logging
import os
import socket
import time
import traceback
import urllib.request, urllib.error, urllib.parse
import jsonschema
import pkg_resources
import psutil
import yaml

from dnstable_manager.download import DownloadManager
from dnstable_manager.fileset import Fileset, FilesetError
from dnstable_manager.util import merge_dicts, TerminatableThread

logger = logging.getLogger(__name__)

class ConfigException(Exception): pass

def get_config(filename=None, stream=None, validate=True):
    configs = []
    with open(pkg_resources.resource_filename(__name__, 'default-config.yaml'), 'r') as f:
        configs.append(yaml.safe_load(f))

    if filename:
        with open(filename, 'r') as f:
            configs.append(yaml.safe_load(f))

    if stream:
        configs.append(yaml.safe_load(stream))

    config = merge_dicts(configs)

    if validate:
        with open(pkg_resources.resource_filename(__name__, 'config-schema.yaml'), 'r') as f:
            schema = yaml.safe_load(f)
        jsonschema.validate(config, schema)

        filesets = set()
        for fileset,fileset_config in list(config['filesets'].items()):
            t = (fileset_config['destination'], fileset_config['base'])
            if t in filesets:
                raise ConfigException('Fileset {} collides with {}/{}.*'.format(fileset, *t))
            filesets.add(t)

            if not os.path.isdir(fileset_config['destination']):
                raise ConfigException('{} is not a directory'.format(fileset_config['destination']))

        for attr in ('ssl_ca_file', 'ssl_keyfile', 'ssl_certfile'):
            if attr in config['downloader']:
                try:
                    open(config['downloader'][attr])
                except IOError as e:
                    raise ConfigException('{attr} {filename}: {strerror}'.format(
                        attr=attr, filename=e.filename, strerror=e.strerror))

    return config

class DNSTableManager:
    def __init__(self, fileset_uri, destination, base=None, extension='mtbl', frequency=1800, download_timeout=None, retry_timeout=60, apikey=None, validator=None, digest_required=True, minimal=True, download_manager=None):
        self.fileset_uri = fileset_uri

        if not os.path.isdir(destination):
            raise OSError(errno.ENOENT, 'Not a directory: \'{}\''.format(destination))

        self.destination = destination

        if base:
            self.base = base
        else:
            self.base = os.path.splitext(os.path.basename(urllib.parse.urlsplit(fileset_uri)[2]))[0]

        self.extension = extension
        self.frequency = frequency
        self.download_timeout = download_timeout
        self.retry_timeout = retry_timeout
        self.minimal = minimal

        self.fileset = Fileset(uri=self.fileset_uri,
                dname=self.destination,
                base=self.base,
                extension=self.extension,
                apikey=apikey,
                validator=validator,
                timeout=download_timeout,
                digest_required=digest_required)

        if download_manager:
            self.download_manager = download_manager
        else:
            self.download_manager = DownloadManager(download_timeout=download_timeout, retry_timeout=retry_timeout)
            self.download_manager.start()

        self.thread = None

    def start(self):
        if self.thread:
            raise Exception

        # Use a TerminatableThread so that unit tests can terminate this thread when the test completes
        self.thread = TerminatableThread(target=self.run)
        self.thread.setDaemon(True)
        self.thread.start()

    def join(self):
        if not self.thread:
            raise Exception

        self.thread.join()
        self.thread = None

    def run(self):
        next_remote_load = 0
        while True:
            now = time.time()
            self.fileset.load_local_fileset()

            try:
                if now >= next_remote_load:
                    self.fileset.load_remote_fileset()
                    next_remote_load = now + self.frequency
            except (FilesetError, urllib.error.URLError, urllib.error.HTTPError, http.client.HTTPException, socket.error) as e:
                logger.error('Failed to load remote fileset {}: {}'.format(self.fileset_uri, str(e)))
                logger.debug(traceback.format_exc())
                next_remote_load = now + self.retry_timeout

            for f in sorted(self.fileset.missing_files(), reverse=True):
                if f not in self.download_manager:
                    self.download_manager.enqueue(f)

            self.fileset.prune_obsolete_files(minimal=self.minimal)
            self.fileset.prune_redundant_files(minimal=self.minimal)

            try:
                self.fileset.write_local_fileset()
                if not self.minimal:
                    self.fileset.write_local_fileset(minimal=False)
            except (IOError, OSError) as e:
                logger.error('Failed to write fileset {}: {}'.format(self.fileset.get_fileset_name(), str(e)))
                logger.debug(traceback.format_exc())

            try:
                self.fileset.purge_deleted_files()
            except OSError as e:
                logger.error('Failed to purge deleted files in {}: {}'.format(self.destination, str(e)))
                logger.debug(traceback.format_exc())

            time.sleep(1)

    def clean_tempfiles(self, unlink_open=False):
        open_files = set()
        for p in psutil.process_iter():
            try:
                try:
                    func = p.get_open_files
                except AttributeError:
                    func = p.open_files

                for f in func():
                    open_files.add(f.path)
            except psutil.AccessDenied:
                pass

        for filename in self.fileset.list_temporary_files():
            if filename in open_files and not unlink_open:
                logger.debug('Not unlinking tempfile {!r}: In use.'.format(filename))
                continue

            logger.debug('Unlinking tempfile: {!r}'.format(filename))
            os.unlink(filename)
