# Copyright (c) 2022 DomainTools LLC
# Copyright (c) 2015-2016, 2019, 2021 by Farsight Security, Inc.
# This work is unpublished, and unauthorized reproduction is prohibited.
# Farsight Security, Inc. owns all rights to this work and intends that
# it be held confidential, preserved as a trade secret, and protected as
# an unpublished copyrighted work.

import base64
import hashlib
import logging

logger = logging.getLogger(__name__)

class DigestError(Exception): pass

class DigestChecker:
    def __init__(self, algorithm, init_chunk=[]):
        logger.debug('algorithm=%s', algorithm)

        self._set_digest_obj(algorithm)
        for c in init_chunk:
            self.update(c)

    def _set_digest_obj(self, algorithm):
        if algorithm is None:
            raise DigestError ('No algorithm provided. Skipping digest check.')
        elif algorithm.lower() in ('sha-224', 'sha224'):
            self.digest_obj = hashlib.sha224()
        elif algorithm.lower() in ('sha-256', 'sha256'):
            self.digest_obj = hashlib.sha256()
        elif algorithm.lower() in ('sha-384', 'sha384'):
            self.digest_obj = hashlib.sha384()
        elif algorithm.lower() in ('sha-512', 'sha512'):
            self.digest_obj = hashlib.sha512()
        else:
            raise DigestError('Unsupported algorithm: {}'.format(algorithm))

    def update(self, chunk):
        self.digest_obj.update(chunk)

    def check(self, digest):
        real_digest = base64.b64encode(self.digest_obj.digest())
        digest_bytes = digest.encode('utf-8')
        if real_digest != digest_bytes:
            raise DigestError('Digest mismatch: {} != {}'.format(real_digest, digest_bytes))


def digest_extension(algorithm):
    if algorithm.lower() in ('sha-224', 'sha224'):
        return 'sha224'
    elif algorithm.lower() in ('sha-256', 'sha256'):
        return 'sha256'
    elif algorithm.lower() in ('sha-384', 'sha384'):
        return 'sha384'
    elif algorithm.lower() in ('sha-512', 'sha512'):
        return 'sha512'
    else:
        raise DigestError('Unknown algorithm: {}'.format(algorithm))

DIGEST_EXTENSIONS = ('sha224', 'sha256', 'sha384', 'sha512')
